/*
 * emdf_value.cpp
 *
 * CEMdFValue class
 *
 * Ulrik Petersen
 * Created: 1/18-2003
 * Last update: 6/23-2003
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#include <emdf_value.h>
#include <string_func.h>
#include <pcre.h>
#include <cassert>
#include <iostream>

// Don't use unless you plan to = assign a new value later
CEMdFValue::CEMdFValue()
{
  // Dummily assign kEVInt kind and 0 value
  m_kind = kEVInt;
  m_u.m_int = 0;
}

CEMdFValue::CEMdFValue(eEVkind kind, long i)
{
  m_kind = kind;
  switch (kind) {
  case kEVInt:
  case kEVEnum:
    m_u.m_int = i;
    break;
  case kEVID_D:
    m_u.m_id_d = i;
    break;
  default:
    assert(false);
  };
}

CEMdFValue::CEMdFValue(const CEMdFValue& other)
{ 
  assign(other);
}

CEMdFValue::CEMdFValue(const std::string& str)
{
  m_kind = kEVString;
  m_u.m_pString = new std::string(str);
}

CEMdFValue::~CEMdFValue()
{
  delete_contents();
}


bool CEMdFValue::compare(const CEMdFValue& other, eComparisonOp op) const
{
  bool bResult;

  switch (m_kind) {
  case kEVInt:
  case kEVEnum:
  case kEVID_D:
    bResult = compare_int(other, op);
    break;
  case kEVString:
    bResult = compare_string(other, op);
    break;
  default:
    assert(false);
  }
  return bResult;
} 

bool CEMdFValue::compare_int(const CEMdFValue& other, eComparisonOp op) const
{
  // Get our own value
  long left_value;
  switch (m_kind) {
  case kEVInt:
    left_value = m_u.m_int;
    break;
  case kEVEnum:
    left_value = m_u.m_int;
    break;
  case kEVID_D:
    left_value = m_u.m_id_d;
    break;
  case kEVString:
    assert(false);
  default:
    assert(false);
  }

  // Get other's value
  long right_value;
  switch (other.m_kind) {
  case kEVInt:
    right_value = other.m_u.m_int;
    break;
  case kEVEnum:
    right_value = other.m_u.m_int;
    break;
  case kEVID_D:
    right_value = other.m_u.m_id_d;
    break;
  case kEVString:
    assert(false);
  default:
    assert(false);
  }

  // Compare
  bool bResult;
  switch (op) {
  case kEqual:
    bResult = left_value == right_value;
    break;
  case kLessThan:
    bResult = left_value < right_value;
    break;
  case kGreaterThan:
    bResult = left_value > right_value;
    break;
  case kNotEqual:
    bResult = left_value != right_value;
    break;
  case kLessThanOrEqual:
    bResult = left_value <= right_value;
    break;
  case kGreaterThanOrEqual:
    bResult = left_value >= right_value;
    break;
  default:
    assert(false);
    break;
  }

  // Return result
  return bResult;
}

bool CEMdFValue::compare_string(const CEMdFValue& other, eComparisonOp op) const
{
  // Make sure that other is string
  if (other.m_kind != kEVString) {
    assert(false);
  }

  // Compare
  bool bResult;
  switch (op) {
  case kEqual:
    bResult = *m_u.m_pString == *other.m_u.m_pString;
    break;
  case kLessThan:
    bResult = *m_u.m_pString < *other.m_u.m_pString;
    break;
  case kGreaterThan:
    bResult = *m_u.m_pString > *other.m_u.m_pString;
    break;
  case kNotEqual:
    bResult = *m_u.m_pString != *other.m_u.m_pString;
    break;
  case kLessThanOrEqual:
    bResult = *m_u.m_pString <= *other.m_u.m_pString;
    break;
  case kGreaterThanOrEqual:
    bResult = *m_u.m_pString >= *other.m_u.m_pString;
    break;
  default:
    assert(false);
    break;
  }

  // Return result
  return bResult;
}

void CEMdFValue::get_as_string(std::string& result) const
{
  switch (m_kind) {
  case kEVInt:
  case kEVEnum:
    result = long2string(m_u.m_int);
    break;
  case kEVID_D:
    result = id_d2string(m_u.m_id_d);
    break;
  case kEVString:
    result = *m_u.m_pString;
    break;
  default:
    assert(false);
  }
}

void CEMdFValue::assign(const CEMdFValue& other)
{
  m_kind = other.m_kind;
  switch (m_kind) {
  case kEVInt:
  case kEVEnum:
    m_u.m_int = other.m_u.m_int;
    break;
  case kEVID_D:
    m_u.m_id_d = other.m_u.m_id_d;
    break;
  case kEVString:
    m_u.m_pString = new std::string(*other.m_u.m_pString);
    break;
  default:
    assert(false);
  };
}

CEMdFValue& CEMdFValue::operator=(const CEMdFValue& other)
{
  delete_contents();
  assign(other);
  return *this;
}

void CEMdFValue::delete_contents()
{
  switch (m_kind) {
  case kEVString:
    delete m_u.m_pString;
    break;
  case kEVInt:
  case kEVEnum:
  case kEVID_D:
    // Do nothing
    break;
  default:
    assert(false);
  };
}



////////////////////////////////////////////////////////////
//
// global functions
//
////////////////////////////////////////////////////////////
std::string eComparisonOp2SQLString(eComparisonOp e)
{
  std::string strResult;
  switch (e) {
  case kEqual:
    strResult = "=";
    break;
  case kLessThan:
    strResult = "<";
    break;
  case kGreaterThan:
    strResult = ">";
    break;
  case kNotEqual:
    strResult = "<>";
    break;
  case kLessThanOrEqual:
    strResult = "<=";
    break;
  case kGreaterThanOrEqual:
    strResult = ">=";
    break;
  default:
    std::cerr << "Error: eComparisonOp2SQLString: eComparisonOp does not have a counterpart in SQL!" << std::endl;
    assert(false);
  }
  return strResult;
}

