//
//
// monads.cpp
// monads and sets of monads
// Ulrik Petersen
// Created: 7/16-1999
// Last update: 6/21-2003
//
//
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 1999-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#include <monads.h>
#include <cassert>
#include <sstream>





//////////////////////////////////////////////////////////////////
//
//  class Monad_Set_Element 
//
//  UTILITY FUNCTIONS
//
//////////////////////////////////////////////////////////////////


std::ostream& operator<<(std::ostream& s, const Monad_Set_Element& mse)
{
  monad_m mse_first = mse.first();
  monad_m mse_last = mse.last();
  if (mse_first == mse_last)
    s << mse_first;
  else
    s << mse_first << "-" << mse_last;
  s << " ";
  return s;
}


/////////////////////////////////////////////////////////////////
//
// SOMConstIterator
//
/////////////////////////////////////////////////////////////////
SOMConstIterator::SOMConstIterator(const CSet_of_monad_ms *pMotherSOM)
  : m_pMotherSOM(pMotherSOM)
{
  m_iter = m_pMotherSOM->monad_ms.begin();
}

SOMConstIterator::SOMConstIterator()
  :  m_pMotherSOM(0)
{
}

SOMConstIterator::SOMConstIterator(const SOMConstIterator& other)
{
  assign(other);
}


SOMConstIterator::~SOMConstIterator()
{
}

SOMConstIterator& SOMConstIterator::operator=(const SOMConstIterator& other)
{
  assign(other);
  return *this;
}

void SOMConstIterator::assign(const SOMConstIterator& other)
{
  m_pMotherSOM = other.m_pMotherSOM;
  m_iter = other.m_iter;
}


bool SOMConstIterator::hasNext() const
{
  assert(m_pMotherSOM != 0);
  return m_iter != m_pMotherSOM->monad_ms.end();
}


Monad_Set_Element SOMConstIterator::next()
{
  assert(m_pMotherSOM != 0);
  assert(hasNext());
  Monad_Set_Element mse = *m_iter;
  ++m_iter;
  return mse;
}

Monad_Set_Element SOMConstIterator::previous()
{
  assert(m_pMotherSOM != 0);
  --m_iter;
  return *m_iter;
}

Monad_Set_Element SOMConstIterator::current()
{
  assert(m_pMotherSOM != 0);
  return *m_iter;
}





//////////////////////////////////////////////////////////////////
//
//  class CSet_of_monad_ms
//
//  A set of monad_m's.
//
//  INVARIANT:
//  For all Monad_Set_Elements mse in the vector, 
//  it is the case that:
//
//     1.1) Either its predecessor prev is empty 
//          (i.e., is not there), OR
//     1.2) prev.last+1 < mse.first,
//          AND
//     2.1) Either its successor succ is empty 
//          (i.e., is not there), OR
//     2.2) mse.last+1 < succ.first 
//
//  so, 
//
//  FOR ALL mse IN vector: (FOR ALL prev before mse: prev.last+1 <
//                        mse.first) AND (FOR ALL succ after mse:
//                        mse.last+1 < succ.first )
//
//  where the two nested FOR ALL quantifiers take care of 1.1 and 2.1 
//  by being vacuously true when prev and succ are not there.
//
//////////////////////////////////////////////////////////////////

CSet_of_monad_ms::CSet_of_monad_ms(const CSet_of_monad_ms& other) 
  : monad_ms(other.monad_ms), m_first(other.m_first), m_last(other.m_last)
{
}

CSet_of_monad_ms::~CSet_of_monad_ms()
{
}



//
// Add a single mse safely.
//
void CSet_of_monad_ms::AddMSE(Monad_Set_Element mse)
{
  // Get mse's first and last
  monad_m mse_first = mse.first_m;
  monad_m mse_last = mse.last_m;

  if (is_empty()) {
    // If we are empty, simply push it at the end
    monad_ms.push_back(mse);
  } else if (mse_first > (m_last+1)) {
    // If it occurs after m_last, simply push it at the end
    monad_ms.push_back(mse);
  }  else {
    // Otherwise, insert carefully...
    Monad_Set_Element current_mse = mse;
    MSEList::iterator cur = monad_ms.begin();

    while (cur != monad_ms.end()) {
      if (current_mse.overlap(*cur) 
	  || current_mse.adjacent(*cur)) {
	current_mse.subsume(*cur);
	cur = monad_ms.erase(cur);
      } else if (cur->past(current_mse)) {
	break;
      } else {
	++cur;
      }
    }
    monad_ms.insert(cur, current_mse);
  }

  // Take care of setting m_first and m_last
  if (mse_first < m_first)
    m_first = mse_first;
  if (mse_last > m_last)
    m_last = mse_last;
}

//
// Put to ostream
// Named so because we don't want it to be ambiguous
// in CObject_dm's whether to call
// CObject_d::put or CSet_of_monad_ms::put.
//
std::ostream& CSet_of_monad_ms::putme(std::ostream& s) const
{
  s << " { ";
  MSEList::const_iterator i = monad_ms.begin();
  MSEList::const_iterator e = monad_ms.end();
  while (i != e) {
    i->putme(s);
    ++i;
    if (i != e)
      s << ", ";
  }
  s << " } ";
  return s;
}

std::string CSet_of_monad_ms::ToString() const
{
  // Just use putme
  std::ostringstream ostr;
  putme(ostr);
  return ostr.str();

  /*
   * The following is actually slower than using putme.
   *
   */
  /*
  out = " { ";
  MSEList::const_iterator i = begin();
  while (i != end()) {
    monad_m mse_first = i->first();
    monad_m mse_last = i->last();
    if (mse_first == mse_last)
      out += monad_m2string(mse_first);
    else
      out += monad_m2string(mse_first) + "-" + monad_m2string(mse_last);
    out += " ";
    ++i;
    if (i != end())
      out += ", ";
  }
  out += "} ";
  */
}


bool CSet_of_monad_ms::operator==(CSet_of_monad_ms& other) const
{
  // Return at once if first/last are not the same
  if (m_first != other.m_first
      || m_last != other.m_last)
    return false;

  return monad_ms == other.monad_ms;
  /*
  MSEList::const_iterator us, them;
  MSEList::const_iterator usend(monad_ms.end()), themend(other.monad_ms.end());

  us = monad_ms.begin();
  them = other.monad_ms.begin();

  while (us != usend && them != themend) {
    if (!(*us == *them)) 
      return false;
    ++us;
    ++them;
  }

  if (us == usend && them == themend)
    return true;
  else
    return false;
  */
}

void CSet_of_monad_ms::get_monad_m_vector(std::vector<monad_m>& result) const
{
  MSEList::const_iterator i(monad_ms.begin());
  MSEList::const_iterator iend(monad_ms.end());
  while (i != iend) {
    monad_m m;
    for (m = i->first_m; m <= i->last_m; ++m) {
      result.push_back(m);
    }
    ++i;
  }
}



CSet_of_monad_ms& CSet_of_monad_ms::operator=(const CSet_of_monad_ms& som)
{
  // Simply copy
  monad_ms = som.monad_ms;

  /*
  // Clear if not empty
  if (!is_empty()) 
    clear();

  MSEList::const_iterator ci(som.begin());
  MSEList::const_iterator cend(som.end());
  while (ci != cend) {
    monad_ms.push_back(*ci);
    ++ci;
  }
  */

  // Set first/last
  m_first = som.m_first;
  m_last = som.m_last;

  // Return ourselves
  return *this;
}

void CSet_of_monad_ms::add(monad_m monad)
{
  Monad_Set_Element mse(monad);
  AddMSE(mse);
}

void CSet_of_monad_ms::add(monad_m first, monad_m last)
{
  Monad_Set_Element mse(first, last);
  AddMSE(mse);
}

// part_of:
// Returns: bool, "Are we part_of other?"
//
// Explanation:
// There are six cases:
//
//               +-------------+
//               |      B      |
//               +-------------+
//    +---+  +-----+  +---+  +-----+  +---+
// A: | 1 |  |  2  |  | 3 |  |  4  |  | 5 |
//    +---+  +-----+  +---+  +-----+  +---+
//          +-----------------------+
//       A: |           6           |
//          +-----------------------+
//
//  We start by comparing the first A to the first B.
//
//  In case 1, we can return because A has monads outside B
//  In case 2, we can return because A has monads outside B
//  In case 4, we can return because A has monads outside B
//  In case 6, we can return because A has monads outside B
//  In case 3, we need to go on to the next A (but not necessarily B,
//             see 5)
//  In case 5, we need to advance B, because we have no decisive 
//             information, though we do not also advance A.
//
//  If we get to the end of A it is because all have been "inside" some
//             MSE in B.  Thus the loop terminates if all are "inside".
//  If we get to the end of B without getting to the end of A, it is 
//             because some element of A was to the right of all elements
//             in B.  Thus A had elements that were not in B.
//
//  Note that the logical and of the negations of all of the first 5
//             implies the sixth.
//
bool CSet_of_monad_ms::part_of(const CSet_of_monad_ms& other) const
{
  if (!((m_first >= other.m_first) && (m_last <= other.m_last)))
    return false;

  // Get iterators
  MSEList::const_iterator A, B;
  MSEList::const_iterator Aend(monad_ms.end()), Bend(other.monad_ms.end());
  A = monad_ms.begin();
  B = other.monad_ms.begin();

  // Get A/B first/last
  monad_m A_first_m, B_first_m;
  monad_m A_last_m, B_last_m;
  A_first_m = A->first_m;
  A_last_m = A->last_m;
  B_first_m = B->first_m;
  B_last_m = B->last_m;

  // Iterate
  while (A != Aend && B != Bend) {
    if (A_last_m < B_first_m)
      return false;                     // 1
    else if (A_first_m < B_first_m 
	     && A_last_m >= B_first_m 
	     && A_last_m <= B_last_m)
      return false;                     // 2
    else if (A_first_m >= B_first_m 
	     && A_last_m <= B_last_m) {
      ++A;                              // 3
      A_first_m = A->first_m;
      A_last_m = A->last_m;
    }
    else if (A_first_m >= B_first_m 
	     && A_first_m <= B_last_m
	     && A_last_m > B_last_m)
      return false;                     // 4
    else if (A_first_m > B_last_m) {
      ++B;                              // 5
      B_first_m = B->first_m;
      B_last_m = B->last_m;
    }
    else
      return false;                     // 6
  }
  if (A == Aend)
    return true;
  else
    return false;
}

// Make this = this union other
void CSet_of_monad_ms::union_with(const CSet_of_monad_ms& other)
{
  MSEList::const_iterator O(other.monad_ms.begin());
  MSEList::const_iterator Oend(other.monad_ms.end());

  while (O != Oend) {
    AddMSE(*O);
    ++O;
  }
}

bool CSet_of_monad_ms::is_member_of(monad_m m) const
{
  MSEList::const_iterator O(monad_ms.begin());
  MSEList::const_iterator Oend(monad_ms.end());

  while (O != Oend) {
    monad_m Ofirst = O->first_m;
    if (m >= Ofirst && m <= O->last_m)
      return true;
    if (m < Ofirst)
      return false;
    ++O;
  }

  return false;
}


//
// Removes the monads of other from us
//
void CSet_of_monad_ms::difference(const CSet_of_monad_ms& other)
{
  MSEList::const_iterator i = other.monad_ms.begin();
  MSEList::const_iterator iend = other.monad_ms.end();
  while (i != iend) {
    RemoveMSE(*i);
    ++i;
  }
}

// RemoveMSE
//
// Removes the monads of an MSE B from ourselves.
//
// Explanation:
// There are six cases:
//
//               +-------------+
//               |      B      |
//               +-------------+
//    +---+  +-----+  +---+  +-----+  +---+
// A: | 1 |  |  2  |  | 3 |  |  4  |  | 5 |
//    +---+  +-----+  +---+  +-----+  +---+
//          +-----------------------+
//       A: |           6           |
//          +-----------------------+
//
// We go through the list of MSEs.
// A is the current one.
// We start at the beginning.
//
//  In case 1, we go on to the next A
//  In case 2, we truncate A: A.last = B.first-1 and go on 
//             to the next one
//  In case 3, we delete A and go on to the next one
//  In case 4, we truncate A: A.first = B.last+1 and go on
//             to the next one
//  In case 5, we break the while-loop, since we are done
//  In case 6, we split A and stop the while-loop, since we are done:
//             oldlast := A.last;  A.last := B.first-1; 
//             insert new A(B.last+1,oldlast) after A;
//             break;
//
//  Note that the logical and of the negations of all of the first 5
//             implies the sixth.
//
void CSet_of_monad_ms::RemoveMSE(const Monad_Set_Element& B)
{
  monad_m B_first = B.first_m;
  monad_m B_last = B.last_m;
  if (is_empty())
    return;
  else {
    MSEList::iterator A;
    A = monad_ms.begin();
    // Get A first/last
    monad_m A_first_m = A->first_m;
    monad_m A_last_m = A->last_m;

    while (A != monad_ms.end()) {
      if (A_last_m < B_first) {     // 1 
	++A;			    
	A_first_m = A->first_m;
	A_last_m = A->last_m;
      } else if (A_first_m < B_first 
	       && A_last_m >= B_first 
	       && A_last_m <= B_last) { // 2
	A->truncate(B);	
	++A;
	A_first_m = A->first_m;
	A_last_m = A->last_m;
      } else if (A_first_m >= B_first 
		 && A_last_m <= B_last) {     // 3
	MSEList::iterator C;
	C = A;
	++C;
	monad_ms.erase(A);
	A = C;
	A_first_m = A->first_m;
	A_last_m = A->last_m;
      } else if (A_first_m >= B_first 
		 && A_first_m <= B_last
		 && A_last_m > B_last) { // 4
	A->truncate(B);                   
	++A;
	A_first_m = A->first_m;
	A_last_m = A->last_m;
      } else if (A_first_m > B_last)	 // 5
	break;                            
      else {                                     // 6
	monad_m oldlast = A_last_m;
	A->truncate(B);
	Monad_Set_Element newmse(B_last+1, oldlast);
	monad_ms.insert(++A, newmse);
	break;
      }
    }
  }

  // Take care of setting m_first and m_last
  if (is_empty()) {
    m_first = MAX_MONAD;
    m_last = 0;
  } else {
    m_first = monad_ms.begin()->first_m;
    m_last = monad_ms.rbegin()->last_m;
  }
}

// Intersect
// Returns the intersection of Aset and Bset.
//
// Explanation:
// There are six cases:
//
//               +-------------+
//               |      B      |
//               +-------------+
//    +---+  +-----+  +---+  +-----+  +---+
// A: | 1 |  |  2  |  | 3 |  |  4  |  | 5 |
//    +---+  +-----+  +---+  +-----+  +---+
//          +-----------------------+
//       A: |           6           |
//          +-----------------------+
//
//  We start by comparing the first A to the first B.
//
//  In case 1, 
//             - Go to the next A
//  In case 2, 
//             - add B.first_m..A.last_m
//             - go to the next A
//  In case 3, 
//             - add A
//             - go to the next A
//  In case 4, 
//             - add A.first_m..B.last_m
//             - go to the next B
//  In case 5, 
//             - go to the next B
//  In case 6, 
//             - add B
//             - go to the next B.
//
//
//  Note that the logical and of the negations of all of the first 5
//             implies the sixth.
//
CSet_of_monad_ms CSet_of_monad_ms::intersect(const CSet_of_monad_ms& Aset, const CSet_of_monad_ms& Bset)
{
  // Declare empty result
  CSet_of_monad_ms result;

  // Get iterators
  MSEList::const_iterator A(Aset.monad_ms.begin());
  MSEList::const_iterator B(Bset.monad_ms.begin());
  MSEList::const_iterator Aend(Aset.monad_ms.end());
  MSEList::const_iterator Bend(Bset.monad_ms.end());

  // Get A/B first/last
  monad_m A_first_m, B_first_m;
  monad_m A_last_m, B_last_m;
  A_first_m = A->first_m;
  A_last_m = A->last_m;
  B_first_m = B->first_m;
  B_last_m = B->last_m;

  // Iterate
  while (A != Aend && B != Bend) {
    if (A_last_m < B_first_m) {      // 1
      //  In case 1, 
      //             - Go to the next A
      ++A;
      A_first_m = A->first_m;
      A_last_m = A->last_m;
    } else if (A_first_m < B_first_m 
	       && A_last_m >= B_first_m 
	       && A_last_m <= B_last_m) { // 2
      //  In case 2, 
      //             - add B.first_m..A.last_m
      //             - go to the next A
      result.add(B_first_m, A_last_m);
      ++A;
      A_first_m = A->first_m;
      A_last_m = A->last_m;
    } else if (A_first_m >= B_first_m 
	       && A_last_m <= B_last_m) { // 3
      //  In case 3, 
      //             - add A
      //             - go to the next A
      result.AddMSE(*A);
      ++A;
      A_first_m = A->first_m;
      A_last_m = A->last_m;
    } else if (A_first_m >= B_first_m 
	       && A_first_m <= B_last_m
	       && A_last_m > B_last_m) { // 4
      //  In case 4, 
      //             - add A.first_m..B.last_m
      //             - go to the next B
      result.add(A->first_m, B->last_m);
      ++B;
      B_first_m = B->first_m;
      B_last_m = B->last_m;
    } else if (A_first_m > B_last_m) {   // 5
      //  In case 5, 
      //               - go to the next B
      ++B;
      B_first_m = B->first_m;
      B_last_m = B->last_m;
    } else {                                   // 6
      //  In case 6, 
      //             - add B
      //             - go to the next B.
      result.AddMSE(*B);
      ++B;
      B_first_m = B->first_m;
      B_last_m = B->last_m;
    }
  }
  return result;
}

//
// Returns the answer to the question:
// "Does a gap exist in ourselves which starts at Sm?
// If the result is true, m will be set to the end of the gap.
//
// The algorithm works by looking at each MSE in the list and
// looking at Sm's position in relation to mse.last_m
//
//    +-----------+
//    |    mse    +
//    +-----------+
//      ^          ^           ^
//    false      true        go on
//   (<last)   (=last+1)   (>last+1)
//
bool CSet_of_monad_ms::gap_exists(monad_m Sm, monad_m& m) const
{
  // The gap must be part_of the set
  if (Sm < m_first || Sm > m_last)
    return false;

  MSEList::const_iterator ci(monad_ms.begin());
  MSEList::const_iterator cend(monad_ms.end());
  while (ci != cend) {
    if (Sm == ci->last_m+1) {
      // Calculate m as next.first_m-1
      // next is always there, because of the if-clause 
      // at the beginning of the function.
      ++ci;
      m = ci->first_m - 1;
      return true;
    } else if (Sm < ci->last_m) {
      return false;
    } 
    ++ci;
  }
  return false;
}

void CSet_of_monad_ms::offset(monad_m m)
{
  MSEList::iterator i(monad_ms.begin());
  MSEList::iterator iend(monad_ms.end());
  while (i != iend) {
    i->offset(m);
    ++i;
  }
  m_first +=m;
  m_last +=m;
}

void CSet_of_monad_ms::clear()
{
  monad_ms.erase(monad_ms.begin(), monad_ms.end());
  m_first = MAX_MONAD;
  m_last = 0;
}


void CSet_of_monad_ms::PrintConsole(CEMdFOutput *pOut) const
{
  pOut->out("{");
  MSEList::const_iterator ci(monad_ms.begin());
  MSEList::const_iterator cend(monad_ms.end());
  while (ci != cend) {
    ci->PrintConsole(pOut);
    ++ci;
    if (ci != cend) {
      pOut->out(",");
    }
  }
  pOut->out(" }");
}

void CSet_of_monad_ms::PrintXML(CEMdFOutput *pOut) const
{
  pOut->start_tag("monad_set", true);
  MSEList::const_iterator ci(monad_ms.begin());
  MSEList::const_iterator cend(monad_ms.end());
  while (ci != cend) {
    ci->PrintXML(pOut);
    ++ci;
  }
  pOut->end_tag("monad_set", true);
}

std::ostream& operator<<(std::ostream& str, const CSet_of_monad_ms& som)
{
  str << " { ";
  SOMConstIterator ci = som.const_iterator();
  while (ci.hasNext()) {
    str << ci.current();
    ci.next();
    if (ci.hasNext())
      str << " , ";
  }
  str << " } ";
  return str;
}

//
// Returns a set of monads which contains all of the monads
// of self, but which has gaps filled in in such as way that the 
// minimum size of any gap in the returned set is min_gap.
CSet_of_monad_ms CSet_of_monad_ms::fill_gaps(monad_m min_gap) const
{
  CSet_of_monad_ms result;

  // Create iterator and previous monad
  MSEList::const_iterator ci;
  monad_m prev_last;

  // Add first MSE
  ci = monad_ms.begin();
  result.AddMSE(*ci);

  // Set previous monad
  prev_last = ci->last_m;

  // Advance ci
  ++ci;

  // Iterate through
  MSEList::const_iterator cend = monad_ms.end();
  while (ci != cend) {

    // Should we add gap or just *ci?
    if ((ci->first_m - prev_last - 1) >= min_gap) {
      // Only add *ci
      result.AddMSE(*ci);
    } else {
      // Add from prev_last to ci->last_m
      result.add(prev_last, ci->last_m);
    }

    // Set previous last monad
    prev_last = ci->last_m;

    // Advance iterator
    ++ci;
  }

  // Return result
  return result;
}

// Get vector with gaps
void CSet_of_monad_ms::get_gap_vector(std::vector<monad_m>& gap_vec) const
{
  // Declare variables
  monad_m prev_last;

  // Get beginning of mselist
  MSEList::const_iterator ci(monad_ms.begin());

  // Get end of mselist
  MSEList::const_iterator cend(monad_ms.end());

  // Set prev_last
  prev_last = ci->last_m;

  // Advance to second mse (if there)
  ++ci;

  // Iterate through
  while (ci != cend) {
    // Get first of this mse
    monad_m this_first = ci->first_m;

    // Push gap to the back
    gap_vec.push_back(this_first - prev_last - 1);

    // Set prev_last
    prev_last = ci->last_m;

    // Advance iterator
    ++ci;
  }
}

SOMConstIterator CSet_of_monad_ms::const_iterator() const
{
  return SOMConstIterator(this);
}



