/*
 * emdros_environment.cpp
 *
 * Emdros environment
 *
 * Ulrik Petersen
 * Created: 3/14-2003
 * Last update: 6/21-2003
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/



#include <emdros-lconfig.h>
#include <emdros_environment.h>
#include <emdf_output.h>
#include <mql_execute.h>
#include <mql_sheaf.h>
#include <mql_error.h>
#include <table.h>
#include <string>
#include <iostream>

#if USE_POSTGRESQL
#include <pgemdfdb.h>
#elif USE_MYSQL
#include <mysqlemdfdb.h>
#endif


// Only used by constructors
void CEmdrosEnv::init(std::ostream* output_stream, 
		      eOutputKind output_kind, eCharsets charset, 
		      std::string hostname, 
		      std::string user, std::string password, 
		      std::string initial_db)
{
  CEMdFOutput *pOut = new CEMdFOutput(charset, output_stream, output_kind);
  
#if USE_POSTGRESQL
  PgEMdFDB *pDB = new PgEMdFDB(hostname, user, password, initial_db);
#elif USE_MYSQL
  MySQLEMdFDB *pDB = new MySQLEMdFDB(hostname, user, password, initial_db);
#endif

  m_pEE = new CMQL_execution_environment(pDB, pOut);
}

CEmdrosEnv::CEmdrosEnv(std::ostream* output_stream, 
		       eOutputKind output_kind, eCharsets charset, 
		       std::string hostname, 
		       std::string user, std::string password, 
		       std::string initial_db)
{
  init(output_stream, output_kind, charset, 
       hostname, user, password, initial_db);
}


// The following constructor uses std::cout (Standard Out)
CEmdrosEnv::CEmdrosEnv(eOutputKind output_kind, eCharsets charset, 
		       std::string hostname, 
		       std::string user, std::string password, 
		       std::string initial_db) 
{
  init(&std::cout, output_kind, charset, 
       hostname, user, password, initial_db);
}


CEmdrosEnv::~CEmdrosEnv()
{
  delete m_pEE->pOut;
  delete m_pEE->pDB;
  delete m_pEE;
}



// Executing queries
bool CEmdrosEnv::execute_string(const std::string& input, bool& bResult, bool bPrintResult, bool bReportError)
{
  return mql_execute_string(m_pEE, input, bResult, bPrintResult, bReportError);
}


bool CEmdrosEnv::execute_file(std::string filename, bool& bResult, bool bPrintResult, bool bReportError)
{
  return mql_execute_file(m_pEE, filename, bResult, bPrintResult, bReportError);
}


bool CEmdrosEnv::execute_stream(std::istream& strin, bool& bResult, bool bPrintResult, bool bReportError)
{
  return mql_execute_stream(m_pEE, strin, bResult, bPrintResult, bReportError);
}


bool CEmdrosEnv::ConnectionOk(void)
{
  return m_pEE->pDB->ConnectionOk();
}

bool CEmdrosEnv::Vacuum(bool bAnalyze)
{
  return m_pEE->pDB->Vacuum(bAnalyze);
}


bool CEmdrosEnv::Get_min_m(monad_m& /* out */ min_m)
{
  return m_pEE->pDB->Get_min_m(min_m);
}


bool CEmdrosEnv::Get_max_m(monad_m& /* out */ max_m)
{
  return m_pEE->pDB->Get_max_m(max_m);
}


bool CEmdrosEnv::Get_all_m_1(CSet_of_monad_ms& /* out */ all_m_1)
{
  return m_pEE->pDB->Get_all_m_1(all_m_1);
}


// Transactions
bool CEmdrosEnv::BeginTransaction(void)
{
  return m_pEE->pDB->BeginTransaction();
}


bool CEmdrosEnv::CommitTransaction(void)
{
  return m_pEE->pDB->CommitTransaction();
}


bool CEmdrosEnv::AbortTransaction(void)
{
  return m_pEE->pDB->AbortTransaction();
}


CMQLResult* CEmdrosEnv::getResult(void)
{
  CStatement *pStatement = m_pEE->pStatement;
  if (pStatement == 0) {
    return 0;
  } else {
    return pStatement->get_result();
  }
}

CSheaf* CEmdrosEnv::getSheaf(void)
{
  CMQLResult *pResult = getResult();
  if (pResult == 0) {
    return 0;
  } else {
    return pResult->get_sheaf();
  }
}

CTable* CEmdrosEnv::getTable(void)
{
  CMQLResult *pResult = getResult();
  if (pResult == 0) {
    return 0;
  } else {
    return (CTable*) pResult;
  }
}


CStatement *CEmdrosEnv::getStatement(void)
{
  return m_pEE->pStatement;

}

CStatement *CEmdrosEnv::takeOverStatement(void)
{
  CStatement *pStatement = m_pEE->pStatement;
  m_pEE->pStatement = 0;
  return pStatement;
}

CMQLResult* CEmdrosEnv::takeOverResult(void)
{
  CStatement *pStatement = takeOverStatement();
  if (!pStatement) {
    return 0;
  } else {
    CMQLResult *pResult = pStatement->get_result();
    pStatement->set_delete_result(false);
    delete pStatement;
    return pResult;
  }
}

CSheaf* CEmdrosEnv::takeOverSheaf(void)
{
  CMQLResult *pMQLResult = takeOverResult();
  if (pMQLResult == 0) {
    return 0;
  } else {
    CSheaf *pResult;
    if (!pMQLResult->isSheaf()) {
      pResult = 0;
      delete pMQLResult;
    } else {
      pResult = pMQLResult->get_sheaf();
      pMQLResult->set_delete_result(false);
      delete pMQLResult;
    }
    return pResult;
  }
}

CTable* CEmdrosEnv::takeOverTable(void)
{
  CMQLResult *pMQLResult = takeOverResult();
  if (pMQLResult == 0) {
    return 0;
  } else {
    CTable *pResult;
    if (!pMQLResult->isTable()) {
      pResult = 0;
      delete pMQLResult;
    } else {
      pResult = pMQLResult->get_table();
    }
    return pResult;
  }
}



// Getting error-messages and info
std::string CEmdrosEnv::get_DB_error(void)
{
  return m_pEE->pDB->ErrorMessage();
}


std::string CEmdrosEnv::get_compiler_error(void)
{
  return m_pEE->pError->mql_get_error();
}


// Gets the compiler stage that was executed last
int CEmdrosEnv::get_last_compiler_stage(void)
{
  return m_pEE->nCompilerStage;
}

void CEmdrosEnv::clear_errors(void)
{
  m_pEE->pDB->clear_local_error();
  m_pEE->pError->mql_clear_error();
}


// Outputting
// These all output on the stream in local CEMdFOutput,
// i.e., the stream you passed to the constructor of CEmdrosEnv,
// or stdout if the other constructor was used
void CEmdrosEnv::Print(std::string str)
{
  m_pEE->pOut->out(str);
}

  // Print your own string
void CEmdrosEnv::Print(CMQLResult *pResult)
{
  pResult->Print(m_pEE->pOut, m_pEE);
}


void CEmdrosEnv::Print(CSheaf *pSheaf)
{
  eOutputKind output_kind = m_pEE->pOut->get_output_kind();
  if (output_kind == kOKConsole) {
    pSheaf->put_Console(m_pEE->pOut, m_pEE);
  } else if (output_kind == kOKXML) {
    pSheaf->put_XML(m_pEE->pOut, m_pEE);
  } else {
    assert(false);
  }
}


void CEmdrosEnv::Print(CTable *pTable)
{
  eOutputKind output_kind = m_pEE->pOut->get_output_kind();
  if (output_kind == kOKConsole) {
    pTable->PrintConsole(m_pEE->pOut);
  } else if (output_kind == kOKXML) {
    pTable->PrintXML(m_pEE->pOut);
  } else {
    assert(false);
  }
}


void CEmdrosEnv::Print(CSet_of_monad_ms *pMonads)
{
  eOutputKind output_kind = m_pEE->pOut->get_output_kind();
  if (output_kind == kOKConsole) {
    pMonads->PrintConsole(m_pEE->pOut);
  } else if (output_kind == kOKXML) {
    pMonads->PrintXML(m_pEE->pOut);
  } else {
    assert(false);
  }
}



//
// XML outputting
//

// XML declaration.
// Calls CEMdFOutput::print_XMLDecl on the local CEMdFOutput
void CEmdrosEnv::Print_XMLDecl()
{
  m_pEE->pOut->print_XMLDecl();
}

 

// DTDs
void CEmdrosEnv::Print_DTD_start(std::string root_element)
{
  m_pEE->pOut->print_DTD_start(root_element);
}


void CEmdrosEnv::Print_DTD_end()
{
  m_pEE->pOut->print_DTD_end();
}


void CEmdrosEnv::Print_DTD_CMQLResult(void)
{
  CMQLResult::print_DTD(m_pEE->pOut);
}


void CEmdrosEnv::Print_DTD_CTable(void)
{
  CTable::print_DTD(m_pEE->pOut);
}


void CEmdrosEnv::Print_DTD_CSheaf(void)
{
  CSheaf::print_DTD(m_pEE->pOut);
}


void CEmdrosEnv::clean()
{
  m_pEE->clean();
}

// Backend-name
std::string CEmdrosEnv::BackendName(void)
{
  return EMdFDB::BackendName();
}

  // Returns the string-representation of an enumeration constant in 
  // enum enum_name with the value value.  If more than one enum constant
  // has the same value, it is undefined which of them is returned.
std::string CEmdrosEnv::GetEnumConstNameFromValue(long value,
						  const std::string& enum_name, 
						  /* out */ bool &bDBOK)
{
  std::string enum_const_name;
  bDBOK = m_pEE->pDB->GetEnumConstNameFromValue(value, enum_name, enum_const_name);
  return enum_const_name;
}
 
bool CEmdrosEnv::isSheaf(void)
{
  CMQLResult *pResult = getResult();
  if (pResult == 0) {
    return false;
  } else {
    return pResult->isSheaf();
  }
}

bool CEmdrosEnv::isTable(void) 
{
  CMQLResult *pResult = getResult();
  if (pResult == 0) {
    return false;
  } else {
    return pResult->isTable();
  }
}
