/*
 * mql_object_type_statements.cpp
 *
 * MQL OBJECT TYPE statements
 *
 * Ulrik Petersen
 * Created: 3/1-2001
 * Last update: 3/20-2003
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#include <mql_object_type_statements.h>
#include <infos.h>
#include <list>


/////////////////////////////////////////////////////////////
//
// Base class for all OBJECT TYPE statements
//
// CObjectTypeStatement
//
/////////////////////////////////////////////////////////////
CObjectTypeStatement::~CObjectTypeStatement()
{
  delete m_object_type_name;
}

//
// Checks whether object type exists.
// Sets m_object_type_id if it does.
bool CObjectTypeStatement::SymbolObjectTypeExists(bool& bResult)
{
  return m_pEE->pDB->ObjectTypeExists(*m_object_type_name,
			       bResult, m_object_type_id);
}

/////////////////////////////////////////////////////////////
//
// CREATE OBJECT TYPE
//
// CCreateObjectTypeStatement 
//
/////////////////////////////////////////////////////////////
CCreateObjectTypeStatement::CCreateObjectTypeStatement(
           CMQL_execution_environment *pEE,
           std::string* object_type_name, 
	   CFeatureDeclaration* feature_declaration_list,
	   CCBBFList* cbbf_list)
  : CObjectTypeStatement(pEE, object_type_name)
{
  m_features = feature_declaration_list;
  m_cbbf_list = cbbf_list;
}

CCreateObjectTypeStatement::~CCreateObjectTypeStatement()
{
  delete m_features;
  delete m_cbbf_list;
}

//
// - Check that "self" is not declared
//
// - Check that declarations of STRING and ASCII have lengths
//   within the prescribed limits.
//
void CCreateObjectTypeStatement::Weed(bool& bResult)
{
  if (m_features) {
    //
    // - Check that "self" is not declared
    //
    // - Check that declarations of STRING and ASCII have lengths
    //   within the prescribed limits.
    //
    m_features->Weed(m_pEE, bResult);
    if (!bResult)
      return;
  }
  else
    bResult = true;
  return;
}

//
// - Check that the object type does not exist
//
// - Check that all the object types in covered_by and
//   buildable_from do exist.
//
// - Check that the enumerations exist for the types 
//   whose features are enumerations
//
// - Check that, within these enumerations, any default
//   specification which is an enumeration constant exists
//   in that enumeration.
//
bool CCreateObjectTypeStatement::Symbol(bool& bResult)
{
  // Check that object type does not exist
  bool bObjectTypeExists;
  if (!SymbolObjectTypeExists(bObjectTypeExists))
    return false;

  if (bObjectTypeExists) {
    m_pEE->pError->mql_append_error("Object type " + *m_object_type_name + " exists already.\n");
    bResult = false;
    return true;
  }

  // Check that all the object types in covered_by 
  // and buildable_from do exist.
  if (m_cbbf_list) {
    if (!m_cbbf_list->SymbolTypesExist(m_pEE, bResult))
      return false;
    if (!bResult) 
      return true;
  }


  if (m_features != 0) {
    //
    // - Check that the enumerations exist for the types   
    //   whose features are enumerations
    if (!m_features->SymbolEnumerationsExist(m_pEE, bResult))
      return false;
    if (!bResult) 
      return true;


    // - Check that, within these enumerations, any default
    //   specification which is an enumeration constant exists
    //   in that enumeration.
    if (!m_features->SymbolEnumConstantsExist(m_pEE, bResult))
      return false;
    if (!bResult)
      return true;
  }


  // Set default
  bResult = true;
  return true;
}

//
// - Assign type-id to each feature.
//
// - Check that there is type-compatibility between the feature
//   and the default specification
//
bool CCreateObjectTypeStatement::Type(bool& bResult)
{
  //
  // - Assign type-id to each feature.
  //
  if (m_features != 0 && !m_features->TypeAssignTypeID(m_pEE)) {
    m_pEE->pError->mql_append_error("Could not assign type id\n");
    return false;
  }

  //
  // - Check that there is type-compatibility between the feature
  //   and the default specification, and make a default specification
  //   if it does not exist.
  //
  if (m_features != 0 && !m_features->TypeTypeCompatibility(m_pEE, bResult)) {
    m_pEE->pError->mql_append_error("Checking type-compatibility failed.\n");
    return false;
  }
  if (!bResult)
    return true;

  // If we got this far, we were a success regardless of bResult's value
  bResult = true;
  return true;
}

bool CCreateObjectTypeStatement::Exec()
{
  // Build list of feature infos
  std::list<feature_info> feature_infos;
  CFeatureDeclaration* f = m_features;
  while (f != 0) {
    // Get feature info from f
    feature_info fi;
    if (!f->get_feature_info(m_pEE, fi))
      return false;

    // Add to list
    feature_infos.push_back(fi);

    // Advance iterator
    f = f->Next();
  }

  // Create object type in DB
  id_d_t object_type_id; // Dummy variable
  if (!m_pEE->pDB->CreateObjectType(*m_object_type_name,
			     feature_infos,
			     object_type_id))
    return false;

  // Add it to m_pEE
  m_pEE->addObjectTypeName(*m_object_type_name);

  // NOTE: We should not do the converse and remove it from m_pEE 
  // when dropping an object type name.
  // This is because m_pEE's list of object types should be strictly
  // monotonic, i.e., one should only add to it.
  // Otherwise, separate queries might get different object type name
  // results, which is possible if you call 
  // CEmdrosEnv::take_over_statement().

  // If we got this far, there were no errors
  return true;
}



/////////////////////////////////////////////////////////////
//
// UPDATE OBJECT TYPE
//
// CUpdateObjectTypeStatement 
//
/////////////////////////////////////////////////////////////
CUpdateObjectTypeStatement::CUpdateObjectTypeStatement(
           CMQL_execution_environment *pEE,
           std::string* object_type_name, 
	   CFeatureUpdate* feature_update_list,
	   CCBBFList* cbbf_list)
  : CObjectTypeStatement(pEE, object_type_name)
{
  m_features = feature_update_list;
  m_cbbf_list = cbbf_list;
}

CUpdateObjectTypeStatement::~CUpdateObjectTypeStatement()
{
  delete m_features;
  delete m_cbbf_list;
}

//
// - Check that "self" is not added or removed
//
// - Check that declarations of STRING and ASCII have lengths
//   within the prescribed limits.
//
void CUpdateObjectTypeStatement::Weed(bool& bResult)
{
  m_features->Weed(m_pEE, bResult);
  return;
}

//
// - Check that the object type already exists.  Store the object 
//   type_id locally
//
// - Check that all the object types in covered_by and
//   buildable_from do exist.  Store the id_d of the covered_by
//   or buildable_from in the AST.
//
// - Check that all the features that are to be removed do exist.
//
// - Check that all the features that are to be added do not exist.
//
// - Check that the enumerations exist for the types 
//   whose features are enumerations for features that are to be
//   added.
//
// - Check that, within these enumerations, any default
//   specification which is an enumeration constant exists
//   in that enumeration for features that are to be added.
//
bool CUpdateObjectTypeStatement::Symbol(bool& bResult)
{
  //
  // - Check that the object type already exists.  Store the object 
  //   type_id locally
  if (!SymbolObjectTypeExists(bResult))
    return false;
  if (!bResult)
    return true;

  // 
  // - Check that all the object types in covered_by and
  //   buildable_from do exist.  Store the id_d of the covered_by
  //   or buildable_from in the AST.
  if (!m_cbbf_list->SymbolTypesExist(m_pEE, bResult))
    return false;
  if (!bResult)
    return true;

  // 
  // - Check that all the features that are to be removed do exist.
  //   
  // - Check that all the features that are to be added do not exist.
  if (!m_features->SymbolExistenceOfFeatures(m_pEE, bResult, m_object_type_id))
    return false;
  if (!bResult)
    return true;

  //
  // - Check that the enumerations exist for the types   
  //   whose features are enumerations
  //
  // - Check that, within these enumerations, any default
  //   specification which is an enumeration constant exists
  //   in that enumeration.
  if (!m_features->SymbolEnumerations(m_pEE, bResult))
    return false;
  if (!bResult)
    return true;

  bResult = true;
  return true;
}

//
// - Assign type-id to each feature.
//
// - Check that there is type-compatibility between the feature
//   and the default specification
//
bool CUpdateObjectTypeStatement::Type(bool& bResult)
{
  //
  // - Assign type-id to each feature.
  //
  if (!m_features->TypeAssignTypeID(m_pEE))
    return false;

  //
  // - Check that there is type-compatibility between the feature
  //   and the default specification
  //
  if (!m_features->TypeTypeCompatibility(m_pEE, bResult))
    return false;

  // If we got this far, we were a success regardless of bResult's value
  return true;
}


bool CUpdateObjectTypeStatement::Exec()
{
  // Start transaction
  bool bDoCommit;
  bDoCommit = m_pEE->pDB->BeginTransaction();

  // Add and remove the features 
  if (!m_features->Exec(m_pEE, *m_object_type_name, m_object_type_id)) {
    if (bDoCommit)
      m_pEE->pDB->AbortTransaction();
    return false;
  }

  // Add and remove the covered_by and buildable_from relations
  if (!m_cbbf_list->Exec(m_pEE, *m_object_type_name, m_object_type_id)) {
    if (bDoCommit)
      m_pEE->pDB->AbortTransaction();
    return false;
  }

  // Commit transaction
  if (bDoCommit) {
    if (!m_pEE->pDB->CommitTransaction()) {
      m_pEE->pError->mql_append_error("CUpdateObjectTypeStatement::Exec: Could not commit transaction.");
      return false;
    }
  }
  return true;
}



/////////////////////////////////////////////////////////////
//
// DROP OBJECT TYPE
//
// CDropObjectTypeStatement 
//
////////////////////////////////////////////////////////////
CDropObjectTypeStatement::CDropObjectTypeStatement(CMQL_execution_environment *pEE, std::string* object_type_name) : CObjectTypeStatement(pEE, object_type_name)
{
}

bool CDropObjectTypeStatement::Symbol(bool& bResult)
{
  //
  // - Check that the object type already exists.  Store the object 
  //   type_id locally
  if (!SymbolObjectTypeExists(bResult))
    return false;

  // In any case, if we got this far, we succeeded.
  return true;
}

bool CDropObjectTypeStatement::Exec()
{
  // Drop object type in DB
  if (!m_pEE->pDB->DropObjectType(*m_object_type_name,
				  m_object_type_id))
    return false;

  // NOTE: We should not remove it from m_pEE 
  // when dropping an object type name.
  // This is because m_pEE's list of object types should be strictly
  // monotonic, i.e., one should only add to it.
  // Otherwise, separate queries might get different object type name
  // results, which is possible if you call 
  // CEmdrosEnv::take_over_statement().

  // If we got this far, there were no errors
  return true;
}



