/*
 * mql_segment_statements.cpp
 *
 * MQL SEGMENT statements
 *
 * Ulrik Petersen
 * Created: 3/6-2001 (March 6, 2001)
 * Last update: 11/22-2002
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2002  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#include <mql_segment_statements.h>


/////////////////////////////////////////////////////////////
//
// CSegmentStatement
//
// base class
//
////////////////////////////////////////////////////////////
CSegmentStatement::~CSegmentStatement()
{
  delete m_segment_name;
}


/////////////////////////////////////////////////////////////
//
// CREATE SEGMENT
//
// CCreateSegmentStatement
//
////////////////////////////////////////////////////////////
CCreateSegmentStatement::CCreateSegmentStatement(
    CMQL_execution_environment *pEE,
    std::string* segment_name, 
    CMonadSetElement* segment_range) : CSegmentStatement(pEE, segment_name),
				       m_segment_range(segment_range)
{

}

CCreateSegmentStatement::~CCreateSegmentStatement()
{
  delete m_segment_range;
}

//
// - Checks that the range is monotonic.
// 
// - Checks that the range is positive
//
void CCreateSegmentStatement::Weed(bool& bResult)
{
  //
  // - Checks that the range is monotonic.
  // 
  m_segment_range->WeedMonotonic(m_pEE, bResult);
  if (!bResult)
    return;

  // 
  // - Checks that the range is positive
  //
  m_segment_range->WeedPositive(m_pEE, bResult);
  return;
}

//
// - Check that the segment does not exist already.
//
// - Check that the new segment does not overlap with any other.
// 
bool CCreateSegmentStatement::Symbol(bool& bResult)
{
  //
  // - Check that the segment does not exist already.
  //
  bool bSegmentExists;
  monad_m dummy_first, dummy_last;
  if (!m_pEE->pDB->SegmentExists(*m_segment_name,
			  bSegmentExists, dummy_first, dummy_last))
    return false;
  if (bSegmentExists) {
    bResult = false;
    std::string mql_error;
    mql_error = "A segment with the name " + *m_segment_name + " already exists.  Cannot create segment of the same name.";
    m_pEE->pError->mql_append_error(mql_error);
    return true;
  }

  //
  // - Check that the new segment does not overlap with any other.
  // 
  if (!m_pEE->pDB->SegmentDoesNotOverlap(m_segment_range->get_mse_first(), 
				  m_segment_range->get_mse_last(), 
				  bResult))
    return false;
  if (!bResult) {
    m_pEE->pError->mql_append_error("New segment overlaps with at least one existing segment.  Cannot create overlapping segments.");
    return true;
  }

  // If we got this far, we are a success.
  bResult = true;
  return true;
}

//
// - Create the segment
//
bool CCreateSegmentStatement::Exec()
{
  //
  // - Create the segment
  //
  if (!m_pEE->pDB->CreateSegment(*m_segment_name,
			  m_segment_range->get_mse_first(), 
			  m_segment_range->get_mse_last()))
    return false;

  // If we got this far, we are a success.
  return true;
}




/////////////////////////////////////////////////////////////
//
// DROP SEGMENT
//
// CDropSegmentStatement
//
////////////////////////////////////////////////////////////
CDropSegmentStatement::CDropSegmentStatement(CMQL_execution_environment *pEE,
std::string* segment_name)
  : CSegmentStatement(pEE, segment_name)
{

}
CDropSegmentStatement::~CDropSegmentStatement()
{

}

//
// - Check that the segment exists already.
//
bool CDropSegmentStatement::Symbol(bool& bResult)
{
  //
  // - Check that the segment exists already.
  //
  bool bSegmentExists;
  monad_m dummy_first, dummy_last;
  if (!m_pEE->pDB->SegmentExists(*m_segment_name,
			  bSegmentExists, dummy_first, dummy_last))
    return false;
  if (!bSegmentExists) {
    bResult = false;
    std::string mql_error;
    mql_error = "No segment with the name " + *m_segment_name + " exists.  Cannot drop nonexisting segment.";
    m_pEE->pError->mql_append_error(mql_error);
    return true;
  }

  // If we got this far, we are a success.
  bResult = true;
  return true;
}

//
// - Drop the segment
// 
bool CDropSegmentStatement::Exec()
{
  //
  // - Drop the segment
  // 
  if (!m_pEE->pDB->DropSegment(*m_segment_name))
    return false;

  // If we got this far, we were a success.
  return true;
}


