/*
 * mql_select_statements.cpp
 *
 * MQL SELECT statements
 *
 * Ulrik Petersen
 * Created: 3/6-2001 (March 6, 2001)
 * Last update: 3/17-2003
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#include <mql_select_statements.h>
#include <mql_query.h>
#include <string_func.h>
#include <mql_R.h>
#include <logging.h>
#include <monads.h>


//////////////////////////////////////////////////////////////
//
// CQueryBase
//
//////////////////////////////////////////////////////////////
CQueryBase::CQueryBase(eFocusSpec focus, 
		       CMonadSetElement* in_spec, 
		       std::string* using_segment,
		       sheaf_return_type_t sheaf_return_type,
		       CTopograph* query) : m_focus(focus),
					    m_in(in_spec),
					    m_using_segment(using_segment),
					    m_sheaf_return_type(sheaf_return_type),

					    m_query(query)
{

}

CQueryBase::~CQueryBase()
{
  delete m_in;
  delete m_using_segment;
  delete m_query;
}

void CQueryBase::MyWeed(CMQL_execution_environment *pEE, bool& bResult)
{
  m_query->Weed(pEE, bResult);
}

bool CQueryBase::MySymbol(CMQL_execution_environment *pEE, bool& bResult)
{
  return m_query->Symbol(pEE, bResult);
}

bool CQueryBase::MyType(CMQL_execution_environment *pEE, bool& bResult)
{
  return m_query->Type(pEE, bResult);
}

bool CQueryBase::MyMonads(CMQL_execution_environment *pEE, bool& bResult)
{
  // Build m_Su
  m_in->MonadsBuildSet(m_Su);

  //
  // Restrict to min_m .. max_m
  //

  // Get all_m-1
  if (!pEE->pDB->Get_all_m_1(pEE->m_all_m_1)) {
    return false;
  }

  // Restrict Su
  m_Su = CSet_of_monad_ms::intersect(m_Su, pEE->m_all_m_1);

  //
  // Restrict to segment
  //
  if (m_using_segment != 0) {
    // Check that segment exists
    monad_m first_m, last_m;
    bool bSegmentExists;
    if (!pEE->pDB->SegmentExists(*m_using_segment, bSegmentExists, first_m, last_m)) {
      return false;
    }
    if (!bSegmentExists) {
      pEE->pError->mql_append_error("Segment '" + *m_using_segment + "' does not exist.");
      bResult = false;
      return true;
    }

    // Create segment
    CSet_of_monad_ms segment(first_m, last_m);

    // Restrict Su
    m_Su = CSet_of_monad_ms::intersect(m_Su, segment);
  }

  // If we got this far, there were no DB errors
  return true;
}

bool CQueryBase::MyExec(CMQL_execution_environment *pEE, CMQLResult **ppMQLResult)
{
  // Calculate multi-maps
  m_query->CalculateMMap();

  // Perform aggregate querying
  if (!m_query->AggregateQuery(pEE, m_Su))
    return false;

  // Declare variables
  CSheaf* pSheaf;

  // Make universe
  CSet_of_monad_ms U(m_Su.first(), m_Su.last());

  // Write to log
  LOG_WRITE_TIME("CQueryBase::MyExec()", "R_topograph starting.");

  // Execute query
  try {
    pSheaf = R_topograph(pEE, U, m_Su, m_query);
  } catch (CEMdFDBException e) {
    return false;
  }

  // Write to log
  LOG_WRITE_TIME("CQueryBase::MyExec()", "R_topograph finished.");

  // Make new CMQLResult from sheaf
  *ppMQLResult = new CMQLResult(pSheaf);

  // If we got this far, there were no DB errors
  return true;
}


/////////////////////////////////////////////////////////////////
//
// SELECT (focus|all|) OBJECTS
//
// CSelectStatement
//
/////////////////////////////////////////////////////////////////
CSelectStatement::CSelectStatement(CMQL_execution_environment *pEE, 
				   eFocusSpec focus, 
				   CMonadSetElement* in_spec, 
				   std::string* using_segment,
				   sheaf_return_type_t sheaf_return_type,
				   CTopograph* query) 
  : CStatement(pEE), CQueryBase(focus, in_spec, using_segment, sheaf_return_type, query)
{
}

CSelectStatement::~CSelectStatement()
{
}

void CSelectStatement::Weed(bool& bResult)
{
  // Just call base-class
  CQueryBase::MyWeed(m_pEE, bResult);
}

bool CSelectStatement::Symbol(bool& bResult)
{
  // Just call base-class
  return CQueryBase::MySymbol(m_pEE, bResult);
}

bool CSelectStatement::Type(bool& bResult)
{
  // Just call base-class
  return CQueryBase::MyType(m_pEE, bResult);
}

bool CSelectStatement::Monads(bool& bResult)
{
  // Just call base-class
  return CQueryBase::MyMonads(m_pEE, bResult);
}

bool CSelectStatement::Exec()
{
  // Just call base-class
  return CQueryBase::MyExec(m_pEE, &m_result);
}

/////////////////////////////////////////////////////////////////
//
// SELECT OBJECTS AT
//
// CSelectObjectsAtStatement
//
/////////////////////////////////////////////////////////////////
CSelectObjectsAtStatement::CSelectObjectsAtStatement(CMQL_execution_environment *pEE, monad_m at_monad, std::string* object_type_name) : CStatement(pEE)
{
  m_at_monad = at_monad;
  m_object_type_name = object_type_name;
  m_object_type_id = NIL;
}

CSelectObjectsAtStatement::~CSelectObjectsAtStatement()
{
  delete m_object_type_name;
}

//
// - Check that the object type exists
//
bool CSelectObjectsAtStatement::Symbol(bool& bResult)
{
  // - Check that the object type exists
  if (!m_pEE->pDB->ObjectTypeExists(*m_object_type_name,
			     bResult,
			     m_object_type_id))
    return false;
  if (!bResult) {
    m_pEE->pError->mql_append_error("Object type name '" + *m_object_type_name + "' does not exist.\n");
    return true;
  }
  
  // If we got this far, there were no DB errors and no compiler errors
  bResult = true;
  return true;
}

//
// - Just ask the database
//
bool CSelectObjectsAtStatement::Exec()
{
  // Get results from DB
  IDDList DBResult;
  if (!m_pEE->pDB->GetObjectsAtM(*m_object_type_name, m_at_monad, DBResult))
    return false;

  // Create new CMQLResult
  m_result = new CMQLResult();
  m_result->append_header("id_d", kTCID_D);
  
  // Add the results from the DB to CMQLResult
  IDDList::const_iterator cbegin(DBResult.begin());
  IDDList::const_iterator ci(cbegin);
  IDDList::const_iterator cend(DBResult.end());
  while (ci != cend) {
    // Add newline unless it is the first
    if (ci != cbegin)
      m_result->newline();

    // Append this id_d
    m_result->append(id_d2string(*ci));

    // Advance iterator
    ++ci;
  }

  // If we got this far, there were no DB errors
  return true;
}


/////////////////////////////////////////////////////////////////
//
// SELECT OBJECTS HAVING MONADS IN
//
// CSelectObjectsHavingMonadsInStatement
//
/////////////////////////////////////////////////////////////////
CSelectObjectsHavingMonadsInStatement::CSelectObjectsHavingMonadsInStatement(CMQL_execution_environment *pEE, 
									     CMonadSetElement* monads, 
									     std::string* object_type_name) : CStatement(pEE), m_AST_monads(monads), m_object_type_name(object_type_name)
{
}

CSelectObjectsHavingMonadsInStatement::~CSelectObjectsHavingMonadsInStatement()
{
  delete m_AST_monads;
  delete m_object_type_name;
}

bool CSelectObjectsHavingMonadsInStatement::Symbol(bool& bResult)
{
  if (strcmp_nocase(*m_object_type_name, "all") != 0) {
    // - Check that the object type exists
    if (!m_pEE->pDB->ObjectTypeExists(*m_object_type_name,
				      bResult,
				      m_object_type_id))
      return false;
    if (!bResult) {
      m_pEE->pError->mql_append_error("Object type name '" + *m_object_type_name + "' does not exist.\n");
      return true;
    }
  }
  
  // If we got this far, there were no DB errors and no compiler errors
  bResult = true;
  return true;
}

// 
// - Check that all the ranges of monads are positive and monotonic
//
void CSelectObjectsHavingMonadsInStatement::Weed(bool& bResult)
{
  // Monad-ranges should be monotonic
  m_AST_monads->WeedMonotonic(m_pEE, bResult);
  if (!bResult)
    return;

  // Monad-ranges should be positive
  m_AST_monads->WeedPositive(m_pEE, bResult);
  if (!bResult)
    return;

  // If we got this far, there were no DB errors and no compiler errors
  bResult = true;
  return;
}

//
// - Build the set of monads
//
bool CSelectObjectsHavingMonadsInStatement::Monads(bool& bResult)
{
  //
  // - Build the set of monads
  //
  m_AST_monads->MonadsBuildSet(m_monads);
  
  // There were no DB errors and no compiler errors
  bResult = true;
  return true;
} 


bool CSelectObjectsHavingMonadsInStatement::Exec()
{
  // Make result
  m_result = new CMQLResult();
  m_result->append_header("object_type_name", kTCString);
  m_result->append_header("monad_m", kTCMonad_m);
  m_result->append_header("id_d", kTCID_D);

  // Are we not doing "all"?
  if (strcmp_nocase(*m_object_type_name, "all") != 0) {
    if (!ExecOneObjectType(*m_object_type_name))
      return false;
  } else {
    // Get object types
    std::list<std::string> object_types;
    if (!m_pEE->pDB->GetObjectTypes(object_types)) {
      return false;
    }

    // Iterate through object types
    std::list<std::string>::const_iterator ci(object_types.begin());
    std::list<std::string>::const_iterator cend(object_types.end());
    while (ci != cend) {
      // Exec for this object type
      if (!ExecOneObjectType(*ci)) {
	m_result->clear();
	return false;
      }

      // Advance iterator
      ++ci;
    }
  } 

  // If we got this far, there were no DB errors
  return true;
}

bool CSelectObjectsHavingMonadsInStatement::ExecOneObjectType(const std::string& object_type_name)
{
  // Ask DB
  if (!m_pEE->pDB->GetObjectsHavingMonadsInMonadSet(object_type_name, m_monads, *m_result)) {
    m_result->clear();
    return false;
  }

  // If we got this far, there were no DB errors
  return true;
}


  



/////////////////////////////////////////////////////////////////
//
// SELECT OBJECT TYPES
//
// CSelectObjectTypesStatement
//
/////////////////////////////////////////////////////////////////
CSelectObjectTypesStatement::CSelectObjectTypesStatement(CMQL_execution_environment *pEE) : CStatement(pEE)
{
}

CSelectObjectTypesStatement::~CSelectObjectTypesStatement()
{
}

bool CSelectObjectTypesStatement::Exec()
{
  // Ask database
  std::list<std::string> l;
  if (!m_pEE->pDB->GetObjectTypes(l)) {
    return false;
  }

  // Create new CMQLResult
  m_result = new CMQLResult();
  m_result->append_header("object type", kTCString);
  
  // Add the results from the DB to CMQLResult
  std::list<std::string>::const_iterator ci(l.begin());
  std::list<std::string>::const_iterator cend(l.end());
  while (ci != cend) {
    m_result->append(*ci);
    ++ci;
    if (ci != cend)
      m_result->newline();
  }

  // Return no database error
  return true;
}



////////////////////////////////////////////////////////////////
//
// SELECT FEATURES
//
// CSelectFeaturesStatement
//
///////////////////////////////////////////////////////////////
//
// - Check that the object type exists
//
bool CSelectFeaturesStatement::Symbol(bool& bResult)
{
  //
  // - Check that the object type exists
  //
  if (!SymbolObjectTypeExists(bResult))
    return false;

  //
  // If we got this far, there were no database errors
  return true;
}

bool CSelectFeaturesStatement::Exec()
{
  // Get result
  std::list<feature_info> feature_infos;
  if (!m_pEE->pDB->GetFeaturesForObjectType(m_object_type_id,
				     feature_infos))
    return false;

  // Create new CMQLResult
  m_result = new CMQLResult();
  m_result->append_header("feature name", kTCString);
  m_result->append_header("type", kTCString);
  m_result->append_header("default", kTCString);
  m_result->append_header("computed", kTCBool);

  // Convert result
  std::list<feature_info>::const_iterator ci(feature_infos.begin());
  std::list<feature_info>::const_iterator cend(feature_infos.end());
  while (ci != cend) {
    // Name
    m_result->append(ci->get_name());

    // Type
    std::string strType_id;
    bool bTypeIDGottenOK = m_pEE->pDB->Type_id_to_string(ci->get_type(), strType_id);
    m_result->append(strType_id);

    // Default value
    std::string defval;
    if (m_pEE->pOut->get_output_kind() == kOKXML) {
      defval = ci->get_default_value();
    } else {
      defval = std::string("'") + ci->get_default_value() + "'";
    }
    m_result->append(defval);

    // Is_computed
    m_result->append(bool_alpha2string(ci->get_is_computed()));
    if (!bTypeIDGottenOK)
      return false;
    ++ci;
    if (ci != cend)
      m_result->newline();
  }
  
  // If we got this far, there were no database errors
  return true;
}


////////////////////////////////////////////////////////////////
//
// SELECT ENUMERATIONS
//
// CSelectEnumerationsStatement
//
///////////////////////////////////////////////////////////////
CSelectEnumerationsStatement::CSelectEnumerationsStatement(CMQL_execution_environment *pEE) : CStatement(pEE)
{
}

CSelectEnumerationsStatement::~CSelectEnumerationsStatement()
{
}

bool CSelectEnumerationsStatement::Exec()
{
  // Ask database
  std::list<std::string> l;
  if (!m_pEE->pDB->GetEnumerations(l)) {
    return false;
  }

  // Create new CMQLResult
  m_result = new CMQLResult();
  m_result->append_header("enum name", kTCString);
  
  // Add the results from the DB to CMQLResult
  std::list<std::string>::const_iterator ci(l.begin());
  std::list<std::string>::const_iterator cend(l.end());
  while (ci != cend) {
    m_result->append(*ci);
    ++ci;
    if (ci != cend)
      m_result->newline();
  }

  // Return no database error
  return true;
}


////////////////////////////////////////////////////////////////
//
// SELECT ENUMERATION CONSTANTS
//
// CSelectEnumerationConstantsStatement
//
///////////////////////////////////////////////////////////////
bool CSelectEnumerationConstantsStatement::Symbol(bool& bResult)
{
  // Check that the enumeration exists and get its enum_id
  if (!SymbolEnumerationExists(bResult)) 
    return false;

  // Return no database error
  return true;
}

bool CSelectEnumerationConstantsStatement::Exec()
{
  // Get result
  std::list<enum_const_info> l;
  if (!m_pEE->pDB->GetEnumConstants(m_enum_id, l))
    return false;

  // Create new CMQLResult
  m_result = new CMQLResult();
  m_result->append_header("enum const", kTCString);
  m_result->append_header("value", kTCInteger);
  m_result->append_header("default", kTCBool);

  // Convert result
  std::list<enum_const_info>::const_iterator ci(l.begin());
  std::list<enum_const_info>::const_iterator cend(l.end());
  while (ci != cend) {
    m_result->append(ci->get_name());
    m_result->append(int2string(ci->get_value()));
    m_result->append(bool_alpha2string(ci->get_is_default()));
    ++ci;
    if (ci != cend)
      m_result->newline();
  }
  
  // If we got this far, there were no database errors
  return true;
}


////////////////////////////////////////////////////////////////
//
// SELECT OBJECT TYPES USING ENUMERATION
//
// CSelectObjectTypesUsingEnumerationStatement
//
///////////////////////////////////////////////////////////////
bool CSelectObjectTypesUsingEnumerationStatement::Symbol(bool& bResult)
{
  // Check that the enumeration exists and get its enum_id
  if (!SymbolEnumerationExists(bResult))
    return false;

  // Return no database error
  return true;
}

bool CSelectObjectTypesUsingEnumerationStatement::Exec()
{
  // Ask database
  std::list<std::string> l;
  if (!m_pEE->pDB->GetObjectTypesUsingEnumeration(m_enum_id, l)) {
    return false;
  }

  // Create new CMQLResult
  m_result = new CMQLResult();
  m_result->append_header("object type name", kTCString);
  
  // Add the results from the DB to CMQLResult
  std::list<std::string>::const_iterator ci(l.begin());
  std::list<std::string>::const_iterator cend(l.end());
  while (ci != cend) {
    m_result->append(*ci);
    ++ci;
    if (ci != cend)
      m_result->newline();
  }

  // Return no database error
  return true;
}


////////////////////////////////////////////////////////////////
//
// SELECT SEGMENTS
//
// CSelectSegmentsStatement
//
///////////////////////////////////////////////////////////////
bool CSelectSegmentsStatement::Exec()
{
  // Ask database
  std::list<segment_info> l;
  if (!m_pEE->pDB->GetSegments(l)) {
    return false;
  }

  // Create new CMQLResult
  m_result = new CMQLResult();
  m_result->append_header("segment name", kTCString);
  m_result->append_header("first monad", kTCMonad_m);
  m_result->append_header("last monad", kTCMonad_m);
  
  // Add the results from the DB to CMQLResult
  std::list<segment_info>::const_iterator ci(l.begin());
  std::list<segment_info>::const_iterator cend(l.end());
  while (ci != cend) {
    m_result->append(ci->get_name());
    m_result->append(monad_m2string(ci->get_first()));
    m_result->append(monad_m2string(ci->get_last()));
    ++ci;
    if (ci != cend)
      m_result->newline();
  }

  // Return no database error
  return true;
}


////////////////////////////////////////////////////////////////
//
// SELECT MAX_M
//
// CSelectMaxMStatement
//
///////////////////////////////////////////////////////////////
bool CSelectMaxMStatement::Exec()
{
  // Ask database
  monad_m max_m;
  if (!m_pEE->pDB->Get_max_m(max_m)) {
    return false;
  }

  // Create new CMQLResult
  m_result = new CMQLResult();
  m_result->append_header("max_m", kTCMonad_m);
  
  // Add the results from the DB to CMQLResult
  m_result->start_new_row();
  m_result->append(monad_m2string(max_m));

  // Return no database error
  return true;
}

////////////////////////////////////////////////////////////////
//
// SELECT MIN_M
//
// CSelectMinMStatement
//
///////////////////////////////////////////////////////////////
bool CSelectMinMStatement::Exec()
{
  // Ask database
  monad_m min_m;
  if (!m_pEE->pDB->Get_min_m(min_m)) {
    return false;
  }

  // Create new CMQLResult
  m_result = new CMQLResult();
  m_result->append_header("min_m", kTCMonad_m);
  
  // Add the results from the DB to CMQLResult
  m_result->start_new_row();
  m_result->append(monad_m2string(min_m));

  // Return no database error
  return true;
}


