/*
 * create_encryption.cpp
 *
 * Program to create encryption.h and default_password.h
 *
 * Ulrik Petersen
 * Created: 1/4-2003 (January 4, 2003)
 * Last update: 6/10-2003
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#include <cstdlib>
#include <cstdio>
#include <ctime>
#include <iostream>
#include <string>


//
// Change this before compilation
//
#define DEFAULT_PASSWORD  ("changeme")

// This should be set to some other number >= 15. The higher the more secure
#define NO_OF_EOR_STRING_CHARS (15)


void seed_random_generator(void)
{
  // Get current time
  time_t curtime = time(NULL);

  // Eor with a number so we don't get -1
  unsigned int seed = ((unsigned int) curtime) ^ 0xed739516;

  // Seed random number generator
  srand(seed);
}

void generate_eor_string(std::string& eor_string)
{
  for (int f = 0; f < NO_OF_EOR_STRING_CHARS; ++f) {
    // Generate something printable
    unsigned char c = 0;
    while (c == 0 
	   || (c < 0x20)
	   || (c >= 0x80 && c <= 0xb0)) {
      c = rand() & 0xFF;
    }

    // Add it to eor_string
    eor_string += c;
  }
}

unsigned int ror(unsigned int i, int rol_bits)
{
  // std::cerr << "Ror in: '" << (unsigned char) i << "'" << std::endl;
  unsigned int a = (i >> rol_bits) & 0xFF;
  unsigned int mask = (1 << rol_bits) - 1;
  int b = (a | ((i & mask) << (8-rol_bits)));
  // std::cerr << "Ror out: '" << (unsigned char) b << "'" << std::endl;
  return b;
}

unsigned char encrypt(unsigned char c, unsigned char prev_char, int rol_bits, const std::string& eor_string)
{
  unsigned int i = c;
  i = (i ^ (unsigned int) prev_char) & 0xFF;
  // std::cerr << "Encrypting '" << c << "'" << std::endl;
  for (int f = eor_string.length() - 1; f >= 0; f--) {
    i = (i ^ (unsigned int) eor_string[f]) & 0xFF;
    if ((f % (eor_string.length()/2)) == 1)
      i = ror(i, rol_bits);
    // std::cerr << "UP1: i = " << i << std::endl;
  }
  // std::cerr << "Result: '" << (unsigned char) i << "'" << std::endl;
  return (unsigned char) i;
}

inline std::string char2hex(unsigned char c)
{
  char szResult[5];
  sprintf(szResult, "\\x%02x", c);
  return szResult;
}

std::string generate_encrypted_password(int rol_bits, int first_eor, const std::string& eor_string, std::string default_password_cleartext)
{
  std::string result;
  unsigned char prev_char = (unsigned char) first_eor;
  for (unsigned int f = 0; f < default_password_cleartext.length(); ++f) {
    result += encrypt(default_password_cleartext[f], prev_char, rol_bits, eor_string);
    prev_char = default_password_cleartext[f];
  }
  return result;
}

#define MAX_LINELENGTH (256+1)

void write_header_to_stdout(int rol_bits, int first_eor, const std::string& hex_eor_string, const std::string& pwd)
{
  char szLine[MAX_LINELENGTH];
  std::cin.getline(szLine, MAX_LINELENGTH-1);
  while (strncmp(szLine, "EOF", 3) != 0) {
    if (strcmp(szLine, ":::PWD:::") == 0) {
      std::cout << "#define PWD       (\"" << pwd << "\")" << std::endl;
    } else if (strcmp(szLine, ":::ROL_BITS:::") == 0) {
      std::cout << "#define ROL_BITS  (" << rol_bits << ")" << std::endl;
    } else if (strcmp(szLine, ":::EOR_STRING:::") == 0) {
      std::cout << "#define EOR_STRING (\"" << hex_eor_string << "\")" << std::endl;
    } else if (strcmp(szLine, ":::FIRST_EOR:::") == 0) {
      std::cout << "#define FIRST_EOR ('" << char2hex(first_eor) << "')" << std::endl;
    } else {
      std::cout << szLine << std::endl;
    }
    std::cin.getline(szLine, MAX_LINELENGTH-1);
  }
}

std::string hexify(const std::string& str)
{
  std::string result;
  for (unsigned int f = 0; f < str.length(); ++f) {
    result += char2hex(str[f]);
  }
  return result;
}

int main(int argc, char *argv[])
{
  // Seed random number generator
  seed_random_generator();

  // Generate the number of rol bits
  int rol_bits = rand() & 0x03;

  // Generate the first eor value
  int first_eor = rand() & 0xFF;

  // Generate eor_string
  std::string eor_string;
  generate_eor_string(eor_string);
  
  // Generate hexified eor_string
  std::string hex_eor_string;
  hex_eor_string = hexify(eor_string);

  // Generate encrypted password
  std::string pwd = "";
  pwd = generate_encrypted_password(rol_bits, first_eor, eor_string, DEFAULT_PASSWORD);

  // Generate hexified pwd
  std::string hex_pwd = "";
  hex_pwd = hexify(pwd);
  
  // Write to stdout
  write_header_to_stdout(rol_bits, first_eor, hex_eor_string, hex_pwd);

  // Return success
  return 0;
}
