//
//
// monads.h
// monad_ms and sets of monads
// Ulrik Petersen
// Created: 7/16-1999
// Last update: 3/16-2004
//
//
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 1999-2004  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/



#ifndef MONADS__H__
#define MONADS__H__

#ifndef SWIG
#include <emdf.h>
#include <emdf_output.h>
#include <vector>
#include <string_func.h>
#endif // !defined SWIG

#undef list
#include <list>
#include <string>

#if HAVE_OSTREAM
#include <ostream>
#else
#include <ostream.h>
#endif

class BadMonadsException {};


class CSet_of_monad_ms; // forward declaration

//////////////////////////////////////////////////////////////////
//
//  class Monad_Set_Element
//
//  Contains two monad_m's, a first and a last.
//  The basic idea is that it represents a contiguous string 
//  of monads, from first till last (inclusive).
//  INVARIANT: first <= last
//
//////////////////////////////////////////////////////////////////
class Monad_Set_Element {
private:
    monad_m first_m, last_m;
public:
    friend class CSet_of_monad_ms; // CSet_of_monad_ms is friend
    Monad_Set_Element(monad_m first, monad_m last) {
      if (first <= last) {
	this->first_m = first;
	this->last_m = last;
      } else {
	throw BadMonadsException();
      }
    };
    Monad_Set_Element(monad_m monad) { first_m = last_m = monad; };

    //
    // Subsume under other.
    // This means:
    //     set first_m to other.first() iff first_m > other.first().
    //     set last_m to other.last() iff last_m < other.last().
    // 
    // If there is no overlap, or if they are not adjacent, a
    // BadMonadsException is thrown.
    //
#ifndef SWIG
    void subsume(Monad_Set_Element& other) {
      if (first_m > other.first_m)
	first_m = other.first_m;
      if (last_m < other.last_m)
	last_m = other.last_m;
    };
    void truncate(const Monad_Set_Element& other) {
      if (first_m < other.first_m)
	last_m = other.first_m-1;
      else if (last_m > other.last_m)
	first_m = other.last_m+1;
    };
    bool adjacent(const Monad_Set_Element& b) const {
      return ((last_m+1 == b.first_m) || (b.last_m+1 == first_m));
    };
    bool overlap(const Monad_Set_Element& b) const {
      return ((last_m >= b.first_m) && (first_m <= b.last_m));
    };
    // Is this.first() > b.last() + 1 ?
    // That is, is this past b?
    bool past(const Monad_Set_Element& b) const {
      return first_m > (b.last_m + 1);
    };
    bool operator==(const Monad_Set_Element& b) const {
      return (first_m == b.first_m && last_m == b.last_m);
    };
#endif // !defined SWIG
    // Returns true on this and b having the same first_m and
    // this and b having the same last_m.
    bool equals(const Monad_Set_Element& b) const {
      return (*this) == b;
    };
    monad_m first(void) const { return first_m; };
    monad_m last(void) const { return last_m; };
    void offset(monad_m m) {
      first_m += m;
      last_m += m;
    };
    void PrintConsole(CEMdFOutput *pOut) const {
      if (first_m == last_m) {
	pOut->out(" " + monad_m2string(first_m));
      } else {
	pOut->out(" " + monad_m2string(first_m) + "-" + monad_m2string(last_m));
      }
    };
    void PrintXML(CEMdFOutput *pOut) const {
      // Make attribute list
      AttributePairList attribute_list;
      attribute_list.push_back(AttributePair("first", monad_m2string(first_m)));
      attribute_list.push_back(AttributePair("last", monad_m2string(last_m)));
      
      // Emit tag
      pOut->start_single_tag("mse", attribute_list, true);
    };

#ifndef SWIG
    std::ostream& putme(std::ostream& s) const {
      if (first_m == last_m)
	s << first_m;
      else
	s << first_m << "-" << last_m;
      return s;
    };
    bool operator<(const Monad_Set_Element& other) const {
      if (first_m < other.first_m)
	return true;
      else if (first_m > other.first_m)
	return false;
      else if (last_m < other.last_m) // So first_m == other.first_m
	return true;
      else 
	return false;
    };
#endif // !defined SWIG
    // Returns true on this object being 
    // "lexicographically" before other object.
    // That is, true if and only if 
    // this->first_m < other.first_m 
    // or (this->first_m == other.first_m 
    //     and
    //     this->last_m < other.last_m)
    bool isBefore(const Monad_Set_Element& other) const {
      return (*this) < other;
    };
};



#ifndef SWIG
typedef std::list<Monad_Set_Element> MSEList;

std::ostream& operator<<(std::ostream& s, const Monad_Set_Element& mse);
#endif

class SOMConstIterator {
private:
  friend class CSet_of_monad_ms;
  const CSet_of_monad_ms *m_pMotherSOM;
  MSEList::const_iterator m_iter;
public:
  SOMConstIterator(const CSet_of_monad_ms *pMotherSOM);
  SOMConstIterator();
  SOMConstIterator(const SOMConstIterator& other);
  ~SOMConstIterator();
#ifndef SWIG
  SOMConstIterator& operator=(const SOMConstIterator& other);
#endif 
  bool hasNext() const; // Is the iterator == end iterator?  Doesn't alter iterator
  Monad_Set_Element next(); // Gets current and advances iterator afterwards
  Monad_Set_Element previous(); // Regresses iterator and then gets current
  Monad_Set_Element current(); // Gets current without altering iterator
private:
  void assign(const SOMConstIterator& other);
};


class CSet_of_monad_ms {
#ifndef SWIG
private:
  friend class SOMConstIterator;
  MSEList monad_ms;
  monad_m m_first, m_last;
#endif // !defined SWIG
public:

  monad_m first() const { return m_first; };
  monad_m last() const { return m_last; };
  SOMConstIterator const_iterator() const;
  std::ostream& putme(std::ostream& s) const;
  std::string ToString(void) const;
  void PrintConsole(CEMdFOutput *pOut) const;
  void PrintXML(CEMdFOutput *pOut) const;
  CSet_of_monad_ms() : m_first(MAX_MONAD), m_last(0) {};
  CSet_of_monad_ms(const CSet_of_monad_ms& other);
  CSet_of_monad_ms(monad_m m) : m_first(m), m_last(m) { monad_ms.push_back(Monad_Set_Element(m)); };
  CSet_of_monad_ms(monad_m first, monad_m last) : m_first(first), m_last(last) { monad_ms.push_back(Monad_Set_Element(first,last)); };
  virtual ~CSet_of_monad_ms();
  bool part_of(const CSet_of_monad_ms& other) const;
  void union_with(const CSet_of_monad_ms& other);
  void difference(const CSet_of_monad_ms& other);
  static CSet_of_monad_ms intersect(const CSet_of_monad_ms& Aset, const CSet_of_monad_ms& Bset);
  void AddMSE(Monad_Set_Element mse);
  void add(monad_m monad);
  void add(monad_m first, monad_m last);
  bool equals(CSet_of_monad_ms& other) const { return *this == other; };
#ifndef SWIG
  CSet_of_monad_ms& operator=(const CSet_of_monad_ms& som);
  bool operator==(CSet_of_monad_ms& other) const;
  void get_monad_m_vector(std::vector<monad_m>& monad_vec) const;
  void get_gap_vector(std::vector<monad_m>& gap_vec) const;
#endif // !defined SWIG
  bool is_member_of(monad_m m) const;
  bool is_empty(void) const { return (monad_ms.begin() == monad_ms.end()); };
  void RemoveMSE(const Monad_Set_Element& mse);
  bool gap_exists(monad_m Sm, monad_m& m) const;
  void offset(monad_m m);
  void clear();
  CSet_of_monad_ms fill_gaps(monad_m min_gap) const;
};

class NoElementsException {};

#ifndef SWIG
std::ostream& operator<<(std::ostream& str, const CSet_of_monad_ms& som);
#endif







#endif //MONADS__H__
