/*
 * pgconn.h
 *
 * PostgreSQL Database connection 
 *
 * Ulrik Petersen
 * Created: 1/27-2001
 * Last update: 6/9-2003
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/




#ifndef EMDF_PGCONN__H__
#define EMDF_PGCONN__H__

extern "C" {
#include <libpq-fe.h>
}
#include <utils.h>
#include <debug.h>
#include <conn.h>
#include <cstdlib>
#include <cassert>

class PgEMdFConnection : public EMdFConnection {
private:
  void *m_pConn;
  void *m_pResult;
  bool m_bTransactionInProgress;
public:
  // Constructor and destructor
  PgEMdFConnection(const std::string& host, const std::string& user, const std::string& passwd, const std::string& database_name);
  virtual ~PgEMdFConnection();

  // Status of connection
  // Returns true on connection good, false on connection bad
  virtual bool ConnectionOk(void);

  // Executing
  virtual bool Exec(const std::string& query);

  // Get last error message
  // Returns true on error message available 
  // (even if it is "OK"), false on error 
  // message not available.
  virtual bool ErrorMessage(std::string& out);

  // Transactions
  virtual bool BeginTransaction(void);
  virtual bool CommitTransaction(void);
  virtual bool AbortTransaction(void);


  // Tuple-counts
  // Both return -1 on not appliccable or error
  virtual long NoOfResultTuples(void);  // SELECT
  virtual long NoOfCommandTuples(void); // INSERT, UPDATE, DELETE

  // Field count
  // Returns -1 on no fields.
  virtual int NoOfFields(void);

  // Tuple-access by field number
  // First tuple is 0
  // First field is 0.
  virtual bool AccessTuple(std::string& out, long tuple_no, int field_no) {
    if (m_pConn == 0)
      return false;
    else {
      const char *szTemp = PQgetvalue((PGresult*) m_pResult, tuple_no, field_no); 
      if (szTemp == NULL) {
	throw EMdFNULLValueException();
      } else {
	out = szTemp;
	return true;
      }
    }
  }
  virtual bool AccessTuple(long tuple_no, 
			   int field_no, 
			   const char* calling_method,
			   bool& result) {
    if (m_pConn == 0) {
      DEBUG_ACCESS_TUPLE_FAILED(calling_method, tuple_no, field_no);
      return false;
    } else {
      const char *szTemp = PQgetvalue((PGresult*) m_pResult, tuple_no, field_no); 
      if (szTemp == NULL) {
	throw EMdFNULLValueException();
      } else {
	// Convert result
	result = char2bool(*szTemp);
	
	// Return success
	return true;
      }
    }
  };
  virtual bool AccessTuple(long tuple_no, 
			   int field_no, 
			   const char* calling_method,
			   long& result) {
    if (m_pConn == 0) {
      DEBUG_ACCESS_TUPLE_FAILED(calling_method, tuple_no, field_no);
      return false;
    } else {
      const char *szTemp = PQgetvalue((PGresult*) m_pResult, tuple_no, field_no); 

      // Convert result
      // NOTE: The following is MUCH slower than strtol or atol.
      // std::string strtmp = szTemp;
      // std::istringstream mystringstream(strtmp);
      // mystringstream >> result;
      
      if (szTemp == NULL) {
	throw EMdFNULLValueException();
      } else {
	// Convert result
	// result = strtol(szTemp, (char **)NULL, 10);
	result = atol(szTemp);

	// Return success
	return true;
      }
    }
  };

  virtual bool AccessTuple(long tuple_no, 
			   int field_no, 
			   const char* calling_method,
			   std::string& result) {

    if (m_pConn == 0) {
      DEBUG_ACCESS_TUPLE_FAILED(calling_method, tuple_no, field_no);
      return false;
    } else {
      const char *szTemp = PQgetvalue((PGresult*) m_pResult, tuple_no, field_no); 


      // Test for result availability      
      if (szTemp == NULL) {
	throw EMdFNULLValueException();
      } else {
	// Convert result
	result = szTemp;

	// Return success
	return true;
      }
    }
  };

  // Release results so they don't take up memory
  virtual void ReleaseResults(void);


  // Vacuum
  virtual bool Vacuum(bool bAnalyze);

private:
  // Make the assignment operator and copy constructor private
  // because they are not supported.
  PgEMdFConnection(const PgEMdFConnection&);
  PgEMdFConnection& operator=(const PgEMdFConnection&);
};


#endif // EMDF_PGCONN__H__
