/*
 * mql.cpp
 *
 * A front-end to the MQL parser
 * Created: 5/1-2001 (1st of May, 2001)
 * Last update: 4/21-2003
 *
 * Return codes:
 * 
 * 0 : Success
 * 1 : Wrong usage
 * 2 : Connection to backend server could not be established
 * 3 : An exception occurred
 * 4 : Could not open file
 * 5 : Database error
 * 6 : Compiler error
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2003  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, license version 2.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 **************************************************************************/


#if HAVE_CONFIG_H
#include <emdros-config.h>
#endif 
#include <emdros-lconfig.h>
#include <emdros_environment.h>
#include <emdfdb.h>
#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#include <default_password.h>

int exec_file(CEmdrosEnv *pEE, std::string filename, bool bPrintOutput)
{
  bool bResult;
  std::string strError;
  std::ifstream fin;
  int nResult;
  fin.open(filename.c_str());
  if (!fin) {
    std::cerr << "FAILURE: Could not open file " << filename << " for reading." << std::endl;
    std::cerr << "Action aborted." << std::endl;
    nResult = 4;
  }
  if (!pEE->execute_stream(fin, bResult, bPrintOutput, true)) {
    std::cerr << "FAILURE: Database error executing " << filename << std::endl;
    nResult = 5;
  } else {
    if (!bResult) {
      // std::cerr << "FAILURE: Compiler error executing " << filename << std::endl;
      nResult = 6;
    } else { 
      // std::cout << "SUCCESS executing file " << filename << std::endl;
      nResult = 0;
    }
  }

  // Return result
  return nResult;
}


int exec_cin(CEmdrosEnv *pEE, bool bPrintOutput)
{
  bool bResult;
  std::string strError;
  int nResult;
  if (!pEE->execute_stream(std::cin, bResult, bPrintOutput, true)) {
    std::cerr << "FAILURE: Database error executing stdin." << std::endl;
    nResult = 5;
  } else {
    if (!bResult) {
      // std::cerr << "FAILURE: Compiler error executing stdin." << std::endl;
      nResult = 6;
    } else { 
      // std::cout << "SUCCESS executing stdin." << std::endl;
      nResult = 0;
    }
  }

  // Return result
  return nResult;
}

void show_version(std::ostream& ostr)
{
  ostr << "mql from Emdros version " << EMDROS_VERSION 
       << " on " << EMdFDB::BackendName() << std::endl;
}

void print_usage(std::ostream& ostr)
{
  show_version(ostr);
  ostr << "Usage:" << std::endl;
  ostr << "   mql [options] [filename] " << std::endl;
  ostr << "       If no filename is given, use stdin." << std::endl;
  ostr << "   mql (--help|-V|--version)" << std::endl;
  ostr << "OPTIONS:" << std::endl;
  ostr << "   --help               Show this help" << std::endl;
  ostr << "   -V , --version       Show version" << std::endl;
  ostr << "   --xml                Output XML format" << std::endl;
  ostr << "   --console            Output console-friendly format" << std::endl;
  ostr << "   -d , --dbname db     Set initial database to db" << std::endl;
  ostr << "   -h , --host host     Set hostname to connect to" << std::endl;
  ostr << "   -u , --user user     Set database user to connect as (default: 'emdf')" << std::endl;
  ostr << "   -p , --password pwd  Set password to use for database user" << std::endl;
  ostr << "   -n , --nop           Do not print results (no output)" << std::endl;
  ostr << "DEFAULTS:" << std::endl;
  ostr << "   --console" << std::endl;
  ostr << "   -d emdf" << std::endl;
  ostr << "   -h localhost" << std::endl;
  ostr << "   -u emdf" << std::endl;
}

bool parse_argument(int argc, char* argv[], int &i, eOutputKind& output_kind, bool& bPrintOutput, std::string& db, std::string& filename, std::string& hostname, std::string& user, std::string& password)
{
  if (argv[i][0] != '-') {
    if (i == (argc-1)) {
      filename = argv[i];
      return true;
    } else {
      // Filename must come at the very end
      std::cerr << "ERROR: filename must be at the very end." << std::endl;
      return false;
    }
  } else { // [0] == '-'
    std::string argument(argv[i]);
    if (argument == "--xml") {
      output_kind = kOKXML;
    } else if (argument == "--console") {
      output_kind = kOKConsole;
    } else if (argument == "--help") {
      print_usage(std::cout);
      exit(0);
    } else if (argument == "-V" || argument == "--version") {
      show_version(std::cout);
      exit(0);
    } else if (argument == "-n" || argument == "--nop") {
      bPrintOutput = false;
    } else if (argument == "-d" || argument == "--dbname") {
      if (i < (argc-1)) {
	++i;
	db = argv[i];
      } else {
	std::cerr << "ERROR: " << argument << " must have a database-name as its argument." << std::endl;
	std::cerr << "       Example: " << argument << " book1" << std::endl;
	return false;
      }
    } else if (argument == "-h"
	       || argument == "--host") {
      if (i < (argc-1)) {
	++i;
	hostname = argv[i];
      } else {
	std::cerr << "ERROR: " << argument << " must have a hostname as its argument." << std::endl;
	std::cerr << "       Example: " << argument << " vlager" << std::endl;
	return false;
      }
    } else if (argument == "-u" || argument == "--user") {
      if (i < (argc-1)) {
	++i;
	user = argv[i];
      } else {
	std::cerr << "ERROR: " << argument << " must have a user as its argument." << std::endl;
	std::cerr << "       Example: " << argument << " emdf" << std::endl;
	return false;
      }
    } else if (argument == "-p" || argument == "--password") {
      if (i < (argc-1)) {
	++i;
	password = argv[i];
      } else {
	std::cerr << "ERROR: " << argument << " must have a password as its argument." << std::endl;
	std::cerr << "       Example: " << argument << " secret" << std::endl;
	return false;
      }
    } else {
      std::cerr << "ERROR: Unrecognized option " << argument << std::endl;
      return false;
    }
    return true;
  }
}

bool parse_arguments(int argc, char* argv[], eOutputKind& output_kind, bool& bPrintOutput, std::string& db, std::string& filename, std::string& hostname, std::string& user, std::string& password)
{
  for (int i=1; i < argc; ++i) {
    if (!parse_argument(argc, argv, i, output_kind, bPrintOutput, db, filename, hostname, user, password)) 
      return false;
  }
  
  return true;

}

int main(int argc, char* argv[])
{
  // Set defaults
  eOutputKind output_kind = kOKConsole;
  std::string initial_db("emdf");
  std::string filename;
  std::string hostname("localhost");
  std::string user("emdf");
  bool bPrintOutput = true;

  // Get default password
  std::string password;
#if USE_DEFAULT_PASSWORD
  get_default_password(password);
#else
  password = "";
#endif

  // Parse arguments
  if (!parse_arguments(argc, argv, output_kind, bPrintOutput, initial_db, filename, hostname, user, password)) {
    print_usage(std::cerr);
    return 1;
  }

  // Make connection
  CEmdrosEnv *pEE = new CEmdrosEnv(output_kind, kCSISO_8859_1,
				   hostname,
				   user, password,
				   initial_db);
  // Zero-fill password
  zero_fill_string(password);

  // Check that we connected
  if (!pEE->ConnectionOk()) {
    std::cerr << "Connection to backend server could not be established." << std::endl;
    std::cerr << "This usually has one of the following causes:" << std::endl;
    std::cerr << "  - The back-end is not running." << std::endl;
    std::cerr << "  - The back-end is not configured for incoming TCP/IP connections." << std::endl;
    std::cerr << "  - You did not supply a user/password with the -u/-p options." << std::endl;
    std::cerr << "    Note that the default user is 'emdf'." << std::endl;
    std::cerr << "  - You have not followed the instructions in the 'bootstrapping.txt'" << std::endl;
    std::cerr << "    document." << std::endl;
    std::cerr << "There are also hints in the HOW-TO-USE document on configuring your back-end." << std::endl;
    std::cerr << "     Please see      <http://emdros.org/whattodo.html>" << std::endl;
    std::cerr << "     for more information on this problem." << std::endl;
    std::cerr << "Also, please be sure to download the full documentation, available from" << std::endl;
    std::cerr << "<http://emdros.org/>.  You won't be able to make Emdros useful without it." << std::endl;
    delete pEE;
    return 2;
  }

  int nResult = 3;
  try {
    if (filename == "") {
      nResult = exec_cin(pEE, bPrintOutput);
    } else {
      nResult = exec_file(pEE, filename, bPrintOutput);
    }
  } catch (CEMdFDBException e) {
    std::cerr << "ERROR: CEMdFDBException (Database error)..." << std::endl;
    std::cerr << pEE->get_DB_error() << std::endl; 
    std::cerr << pEE->get_compiler_error() << std::endl;
  } catch (BadMonadsException e) {
    std::cerr << "BadMonadsException caught.  Program aborted." << std::endl;
  } catch (NoElementsException e) {
    std::cerr << "NoElementsException caught.  Program aborted." << std::endl;
  } catch (CMQLSymbolTableEntryException e) {
    std::cerr << "CMQLSymbolTableEntryException caught.  Program aborted." << std::endl;
  } catch (CMQLSymbolTableException e) {
    std::cerr << "CMQLSymbolTableException caught.  Program aborted." << std::endl;
  } catch (CEMdFOutputWrongCharactersetException e) {
    std::cerr << "CEMdFOutputWrongCharactersetException caught.  Program aborted." << std::endl;
  } catch (CEMdFOutputException e) {
    std::cerr << "CEMdFOutputException caught.  Program aborted." << std::endl;
  } catch (...) {
    std::cerr << "Unknown exception occurred.  Program aborted." << std::endl;
  }

  // Clean up
  delete pEE;
  return nResult;
}


